from dataclasses import dataclass
from enum import Enum
from typing import  Union, Optional

from pychemstation.utils.tray_types import Tray


class SourceType(Enum):
    DEFAULT = "ActualPosition"
    SPECIFIC_LOCATION = "ActualPositionPlusLocation"
    LOCATION = "Location"


class Mode(Enum):
    DEFAULT = "Default"
    SET = "Set"


@dataclass
class Draw:
    amount: Optional[float] = None
    location: Optional[str] = None
    source: Optional[Tray] = None


@dataclass
class Wait:
    duration: int


@dataclass
class Inject:
    pass


class RemoteCommand(Enum):
    START = "START"
    PREPARE = "PREPARE"


@dataclass
class Remote:
    command: RemoteCommand
    duration: int


InjectorFunction = Union[Draw, Wait, Inject, Remote]


@dataclass
class InjectorTable:
    functions: list[InjectorFunction]



import os
import time

from xsdata.formats.dataclass.parsers import XmlParser

from .table import TableController
from ..devices.injector import InjectorController
from ....control.controllers import CommunicationController
from ....generated import PumpMethod, DadMethod, SolventElement
from ....utils.chromatogram import TIME_FORMAT, AgilentChannelChromatogramData
from ....utils.macro import *
from ....utils.method_types import *
from ....utils.table_types import *


class MethodController(TableController):
    """
    Class containing method related logic
    """

    def __init__(self, controller: CommunicationController,
                 src: str,
                 data_dir: str,
                 table: Table,
                 offline: bool,
                 injector_controller: InjectorController):
        self.injector_controller = injector_controller
        super().__init__(controller, src, data_dir, table, offline=offline)

    def check(self) -> str:
        time.sleep(2)
        self.send(Command.GET_METHOD_CMD)
        time.sleep(2)
        res = self.receive()
        if res.is_ok():
            return res.ok_value.string_response
        return "ERROR"

    def get_method_params(self) -> HPLCMethodParams:
        return HPLCMethodParams(organic_modifier=self.controller.get_num_val(
            cmd=TableOperation.GET_OBJ_HDR_VAL.value.format(
                register=self.table.register,
                register_flag=RegisterFlag.SOLVENT_B_COMPOSITION
            )
        ),
            flow=self.controller.get_num_val(
                cmd=TableOperation.GET_OBJ_HDR_VAL.value.format(
                    register=self.table.register,
                    register_flag=RegisterFlag.FLOW
                )
            ),
        )

    def get_row(self, row: int) -> TimeTableEntry:
        flow = None
        om = None

        try:
            flow = self.get_num(row, RegisterFlag.TIMETABLE_FLOW)
        except RuntimeError:
            pass
        try:
            om = self.get_num(row, RegisterFlag.TIMETABLE_SOLVENT_B_COMPOSITION)
        except RuntimeError:
            pass

        return TimeTableEntry(start_time=self.get_num(row, RegisterFlag.TIME),
                              organic_modifer=om,
                              flow=flow)

    def get_timetable(self, rows: int):
        uncoalesced_timetable_rows = [self.get_row(r + 1) for r in range(rows)]
        timetable_rows = {}
        for row in uncoalesced_timetable_rows:
            time_key = str(row.start_time)
            if time_key not in timetable_rows.keys():
                timetable_rows[time_key] = TimeTableEntry(start_time=row.start_time,
                                                          flow=row.flow,
                                                          organic_modifer=row.organic_modifer)
            else:
                if row.flow:
                    timetable_rows[time_key].flow = row.flow
                if row.organic_modifer:
                    timetable_rows[time_key].organic_modifer = row.organic_modifer
        entries = list(timetable_rows.values())
        entries.sort(key=lambda e: e.start_time)
        return entries

    def load(self) -> MethodDetails:
        rows = self.get_num_rows()
        if rows.is_ok():
            self.send(Command.GET_METHOD_CMD)
            res = self.receive()
            method_name = res.ok_value.string_response
            timetable_rows = self.get_timetable(int(rows.ok_value.num_response))
            params = self.get_method_params()
            stop_time = self.controller.get_num_val(
                cmd=TableOperation.GET_OBJ_HDR_VAL.value.format(
                    register=self.table.register,
                    register_flag=RegisterFlag.MAX_TIME))
            post_time = self.controller.get_num_val(
                cmd=TableOperation.GET_OBJ_HDR_VAL.value.format(
                    register=self.table.register,
                    register_flag=RegisterFlag.POST_TIME))
            self.table_state = MethodDetails(name=method_name,
                                             timetable=timetable_rows,
                                             stop_time=stop_time,
                                             post_time=post_time,
                                             params=params)
            return self.table_state
        else:
            raise RuntimeError(rows.err_value)

    def current_method(self, method_name: str):
        """
        Checks if a given method is already loaded into Chemstation. Method name does not need the ".M" extension.

        :param method_name: a Chemstation method
        :return: True if method is already loaded
        """
        self.send(Command.GET_METHOD_CMD)
        parsed_response = self.receive()
        return method_name in parsed_response

    def switch(self, method_name: str):
        """
        Allows the user to switch between pre-programmed methods. No need to append '.M'
        to the end of the method name. For example. for the method named 'General-Poroshell.M',
        only 'General-Poroshell' is needed.

        :param method_name: any available method in Chemstation method directory
        :raise IndexError: Response did not have expected format. Try again.
        :raise AssertionError: The desired method is not selected. Try again.
        """
        self.send(Command.SWITCH_METHOD_CMD_SPECIFIC.value.format(method_dir=self.src,
                                                         method_name=method_name))

        time.sleep(2)
        self.send(Command.GET_METHOD_CMD)
        time.sleep(2)
        res = self.receive()
        if res.is_ok():
            parsed_response = res.ok_value.string_response
            assert parsed_response == f"{method_name}.M", "Switching Methods failed."
        self.table_state = None

    def load_from_disk(self, method_name: str) -> MethodDetails:
        """
        Retrieve method details of an existing method. Don't need to append ".M" to the end. This assumes the
        organic modifier is in Channel B and that Channel A contains the aq layer. Additionally, assumes
         only two solvents are being used.

        :param method_name: name of method to load details of
        :raises FileNotFoundError: Method does not exist
        :return: method details
        """
        method_folder = f"{method_name}.M"
        method_path = os.path.join(self.src, method_folder, "AgilentPumpDriver1.RapidControl.MethodXML.xml")
        dad_path = os.path.join(self.src, method_folder, "Agilent1200erDadDriver1.RapidControl.MethodXML.xml")

        if os.path.exists(os.path.join(self.src, f"{method_name}.M")):
            parser = XmlParser()
            method = parser.parse(method_path, PumpMethod)
            dad = parser.parse(dad_path, DadMethod)

            organic_modifier: Optional[SolventElement] = None
            aq_modifier: Optional[SolventElement] = None

            if len(method.solvent_composition.solvent_element) == 2:
                for solvent in method.solvent_composition.solvent_element:
                    if solvent.channel == "Channel_A":
                        aq_modifier = solvent
                    elif solvent.channel == "Channel_B":
                        organic_modifier = solvent

            self.table_state = MethodDetails(name=method_name,
                                             params=HPLCMethodParams(organic_modifier=organic_modifier.percentage,
                                                                     flow=method.flow),
                                             stop_time=method.stop_time.stop_time_value,
                                             post_time=method.post_time.post_time_value,
                                             timetable=[TimeTableEntry(start_time=tte.time,
                                                                       organic_modifer=tte.percent_b,
                                                                       flow=method.flow
                                                                       ) for tte in method.timetable.timetable_entry],
                                             dad_wavelengthes=dad.signals.signal)
            return self.table_state
        else:
            raise FileNotFoundError

    def edit(self, updated_method: MethodDetails, save: bool):
        """Updated the currently loaded method in ChemStation with provided values.

        :param updated_method: the method with updated values, to be sent to Chemstation to modify the currently loaded method.
        :param save: if false only modifies the method, otherwise saves to disk
        """
        self.table_state = updated_method
        initial_organic_modifier: Param = Param(val=updated_method.params.organic_modifier,
                                                chemstation_key=RegisterFlag.SOLVENT_B_COMPOSITION,
                                                ptype=PType.NUM)
        max_time: Param = Param(val=updated_method.stop_time,
                                chemstation_key=RegisterFlag.MAX_TIME,
                                ptype=PType.NUM)
        post_time: Param = Param(val=updated_method.post_time,
                                 chemstation_key=RegisterFlag.POST_TIME,
                                 ptype=PType.NUM)
        flow: Param = Param(val=updated_method.params.flow,
                            chemstation_key=RegisterFlag.FLOW,
                            ptype=PType.NUM)

        # Method settings required for all runs
        self.update_method_params(flow, initial_organic_modifier, max_time, post_time)
        self._update_method_timetable(updated_method.timetable)

        if save:
            self.send(Command.SAVE_METHOD_CMD.value.format(
                commit_msg=f"saved method at {str(time.time())}"
            ))

    def update_method_params(self, flow, initial_organic_modifier, max_time, post_time):
        self.delete_table()
        self._update_param(initial_organic_modifier)
        self._update_param(flow)
        if self.table_state.stop_time:
            self._update_param(Param(val="Set", chemstation_key=RegisterFlag.STOPTIME_MODE, ptype=PType.STR))
            self._update_param(max_time)
        else:
            self._update_param(Param(val="Off", chemstation_key=RegisterFlag.STOPTIME_MODE, ptype=PType.STR))
        if self.table_state.post_time:
            self._update_param(Param(val="Set", chemstation_key=RegisterFlag.POSTIME_MODE, ptype=PType.STR))
            self._update_param(post_time)
        else:
            self._update_param(Param(val="Off", chemstation_key=RegisterFlag.POSTIME_MODE, ptype=PType.STR))
        self.download()

    def _update_param(self, method_param: Param):
        """Change a method parameter, changes what is visibly seen in Chemstation GUI.
         (changes the first row in the timetable)

        :param method_param: a parameter to update for currently loaded method.
        """
        register = self.table.register
        setting_command = TableOperation.UPDATE_OBJ_HDR_VAL if method_param.ptype == PType.NUM else TableOperation.UPDATE_OBJ_HDR_TEXT
        if isinstance(method_param.chemstation_key, list):
            for register_flag in method_param.chemstation_key:
                self.send(setting_command.value.format(register=register,
                                                       register_flag=register_flag,
                                                       val=method_param.val))
        else:
            self.send(setting_command.value.format(register=register,
                                                   register_flag=method_param.chemstation_key,
                                                   val=method_param.val))
        time.sleep(2)

    def download(self):
        self.send('Sleep 1')
        self.sleepy_send("DownloadRCMethod PMP1")
        self.send('Sleep 1')

    def edit_row(self, row: TimeTableEntry, first_row: bool = False):
        if first_row:
            if row.organic_modifer:
                self.add_row()
                self.add_new_col_text(col_name=RegisterFlag.FUNCTION, val=RegisterFlag.SOLVENT_COMPOSITION.value)
                self.add_new_col_num(col_name=RegisterFlag.TIME, val=row.start_time)
                self.add_new_col_num(col_name=RegisterFlag.TIMETABLE_SOLVENT_B_COMPOSITION, val=row.organic_modifer)
            if row.flow:
                self.add_row()
                self.get_num_rows()
                self.edit_row_text(col_name=RegisterFlag.FUNCTION, val=RegisterFlag.FLOW.value)
                self.add_new_col_num(col_name=RegisterFlag.TIMETABLE_FLOW, val=row.flow)
                self.edit_row_num(col_name=RegisterFlag.TIMETABLE_FLOW, val=row.flow)
            self.download()
        else:
            if row.organic_modifer:
                self.add_row()
                self.get_num_rows()
                self.edit_row_text(col_name=RegisterFlag.FUNCTION, val=RegisterFlag.SOLVENT_COMPOSITION.value)
                self.edit_row_num(col_name=RegisterFlag.TIME, val=row.start_time)
                self.edit_row_num(col_name=RegisterFlag.TIMETABLE_SOLVENT_B_COMPOSITION, val=row.organic_modifer)
                self.download()
            if row.flow:
                self.add_row()
                self.get_num_rows()
                self.edit_row_text(col_name=RegisterFlag.FUNCTION, val=RegisterFlag.FLOW.value)
                self.edit_row_num(col_name=RegisterFlag.TIMETABLE_FLOW, val=row.flow)
                self.edit_row_num(col_name=RegisterFlag.TIME, val=row.start_time)
                self.download()

    def _update_method_timetable(self, timetable_rows: list[TimeTableEntry]):
        self.get_num_rows()

        self.delete_table()
        res = self.get_num_rows()
        while not res.is_err():
            self.delete_table()
            res = self.get_num_rows()

        self.new_table()
        self.get_num_rows()

        for i, row in enumerate(timetable_rows):
            self.edit_row(row=row, first_row=i == 0)

    def stop(self):
        """
        Stops the method run. A dialog window will pop up and manual intervention may be required.\
        """
        self.send(Command.STOP_METHOD_CMD)

    def run(self, experiment_name: str, stall_while_running: bool = True):
        """
        This is the preferred method to trigger a run.
        Starts the currently selected method, storing data
        under the <data_dir>/<experiment_name>.D folder.
        The <experiment_name> will be appended with a timestamp in the '%Y-%m-%d-%H-%M' format.
        Device must be ready.

        :param experiment_name: Name of the experiment
        """

        folder_name = ""
        hplc_is_running = False
        tries = 0
        while tries < 10 and not hplc_is_running:
            timestamp = time.strftime(TIME_FORMAT)
            self.send(Command.RUN_METHOD_CMD.value.format(data_dir=self.data_dir,
                                                          experiment_name=experiment_name,
                                                          timestamp=timestamp))
            folder_name = f"{experiment_name}_{timestamp}.D"
            hplc_is_running = self.check_hplc_is_running()
            tries += 1

        if not hplc_is_running:
            raise RuntimeError("Method failed to start.")

        self.data_files.append(os.path.join(self.data_dir, folder_name))

        if stall_while_running:

            if not self.table_state:
                self.table_state = self.load()

            run_completed = self.check_hplc_done_running(method=self.table_state)
            if run_completed.is_ok():
                self.data_files[-1] = run_completed.ok_value
            else:
                raise RuntimeError("Run error has occurred.")
        else:
            self.data_files[-1] = self.fuzzy_match_most_recent_folder(folder_name).ok_value

    def retrieve_recent_data_files(self) -> str:
        return self.data_files[-1]

    def get_data(self, custom_path: Optional[str] = None,
                 read_uv: bool = False) -> AgilentChannelChromatogramData:
        if not custom_path:
            self.get_spectrum(self.data_files[-1], read_uv)
        else:
            self.get_spectrum(custom_path, read_uv)
        return AgilentChannelChromatogramData(**self.spectra) if not read_uv else self.uv



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Union, Any, Optional

from .injector_types import InjectorTable
from .table_types import RegisterFlag
from ..generated import Signal


class PType(Enum):
    STR = "str"
    NUM = "num"


@dataclass
class Param:
    ptype: PType
    val: Union[float, int, str, Any]
    chemstation_key: Union[RegisterFlag, list[RegisterFlag]]


@dataclass
class HPLCMethodParams:
    organic_modifier: float
    flow: float
    pressure: Optional[float] = None  # TODO: find this


@dataclass
class TimeTableEntry:
    start_time: float
    organic_modifer: float
    flow: Optional[float] = None


@dataclass
class MethodDetails:
    """An Agilent Chemstation method, TODO is to add MS parameters, injector parameters

    :attribute name: the name of the method, should be the same as the Chemstation method name.
    :attribute timetable: list of entries in the method timetable
    :attribute stop_time: the time the method stops running after the last timetable entry.
    :attribute post_time: the amount of time after the stoptime that the pumps keep running,
        based on settings in the first row of the timetable.
    :attribute params: the organic modifier (pump B) and flow rate displayed for the method (the time 0 settings)
    :attribute dad_wavelengthes:
    """
    name: str
    params: HPLCMethodParams
    timetable: list[TimeTableEntry]
    injector_program: Optional[InjectorTable] = None
    stop_time: Optional[float] = None
    post_time: Optional[float] = None
    dad_wavelengthes: Optional[list[Signal]] = None



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

from pychemstation.utils.tray_types import TenVialColumn, Tray


@dataclass
class SequenceDataFiles:
    sequence_name: str
    dir: str
    child_dirs: Optional[list[str]] = None


class SampleType(Enum):
    SAMPLE = 1
    BLANK = 2
    CALIBRATION = 3
    CONTROL = 4

    @classmethod
    def _missing_(cls, value):
        return cls.SAMPLE


class InjectionSource(Enum):
    AS_METHOD = "As Method"
    MANUAL = "Manual"
    MSD = "MSD"
    HIP_ALS = "HipAls"

    @classmethod
    def _missing_(cls, value):
        return cls.HIP_ALS


@dataclass
class SequenceEntry:
    sample_name: str
    vial_location: Tray
    data_file: Optional[str] = None
    method: Optional[str] = None
    num_inj: Optional[int] = 1
    inj_vol: Optional[float] = 2
    inj_source: Optional[InjectionSource] = InjectionSource.HIP_ALS
    sample_type: Optional[SampleType] = SampleType.SAMPLE


@dataclass
class SequenceTable:
    name: str
    rows: list[SequenceEntry]




import unittest

from pychemstation.utils.tray_types import FiftyFourVialPlate, Plate, Letter, Num
from tests.constants import *

run_too = True


class TestRunsStable(unittest.TestCase):
    def setUp(self):
        self.hplc_controller = set_up_utils(254)

    def test_run_method(self):
        try:
            self.hplc_controller.run_method(experiment_name="test_experiment")
            chrom = self.hplc_controller.get_last_run_method_data()
        except Exception as e:
            self.fail(f"Should have not failed: {e}")

    def test_run_10_times(self):
        self.hplc_controller.method_controller.switch(DEFAULT_METHOD)
        rand_method = MethodDetails(
            name=DEFAULT_METHOD,
            params=HPLCMethodParams(
                organic_modifier=5,
                flow=0.65),
            timetable=[TimeTableEntry(
                start_time=0.50,
                organic_modifer=99,
                flow=0.65)],
            stop_time=1,
            post_time=0)
        self.hplc_controller.edit_method(rand_method, save=True)
        try:
            for _ in range(10):
                self.hplc_controller.run_method(experiment_name="limiting_testing")
        except Exception as e:
            self.fail(f"Should have not failed: {e}")

    def test_update_method_update_seq_table_run(self):
        try:
            loc = FiftyFourVialPlate(plate=Plate.ONE, letter=Letter.F, num=Num.TWO)
            self.hplc_controller.switch_sequence(sequence_name=DEFAULT_SEQUENCE)
            seq_table = SequenceTable(
                name=DEFAULT_SEQUENCE,
                rows=[SequenceEntry(
                    vial_location=loc,
                    sample_name="run seq with new method",
                    method=DEFAULT_METHOD,
                    inj_source=InjectionSource.HIP_ALS,
                    inj_vol=0.5,
                    num_inj=1,
                    sample_type=SampleType.SAMPLE
                )])
            self.hplc_controller.edit_sequence(seq_table) # bug didnt delete all the rows, left an extra

            self.hplc_controller.method_controller.switch(DEFAULT_METHOD)
            rand_method = MethodDetails(
                name=DEFAULT_METHOD,
                params=HPLCMethodParams(
                    organic_modifier=5,
                    flow=0.65),
                timetable=[TimeTableEntry(
                    start_time=random.randint(1, 3) + 0.50,
                    organic_modifer=100,
                    flow=0.65)],
                stop_time=random.randint(4, 6),
                post_time=1)
            self.hplc_controller.edit_method(rand_method, save=True)
            if run_too:
                self.hplc_controller.preprun()
                self.hplc_controller.run_sequence()
                chrom = self.hplc_controller.get_last_run_sequence_data()
                # report = process_csv_report(self.hplc_controller.sequence_controller.data_files[-1].child_dirs[-1])
                # self.assertEqual(loc, report.ok_value.vial_location)
        except Exception:
            self.fail("Failed")

    def test_run_sequence(self):
        try:
            self.hplc_controller.switch_sequence(sequence_name=DEFAULT_SEQUENCE)
            self.hplc_controller.preprun()
            self.hplc_controller.run_sequence()
            chrom = self.hplc_controller.get_last_run_sequence_data()
            self.assertTrue(len(chrom) == 1)
        except Exception:
            self.fail("Failed")



from __future__ import annotations

import math
from dataclasses import dataclass
from enum import Enum
from typing import Union


class Num(Enum):
    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    SIX = 6
    SEVEN = 7
    EIGHT = 8
    NINE = 9

    @classmethod
    def from_num(cls, num: int) -> Num:
        match num:
            case 1:
                return Num.ONE
            case 2:
                return Num.TWO
            case 3:
                return Num.THREE
            case 4:
                return Num.FOUR
            case 5:
                return Num.FIVE
            case 6:
                return Num.SIX
            case 7:
                return Num.SEVEN
            case 8:
                return Num.EIGHT
            case 9:
                return Num.NINE
            case _:
                raise ValueError("Num is one of 1 to 9")


class Plate(Enum):
    ONE = -96
    TWO = 4000

    @classmethod
    def from_num(cls, plate: int) -> Plate:
        if 1 <= plate <= 2:
            return Plate.ONE if plate == 1 else Plate.TWO
        raise ValueError("Plate is one or 1 or 2")


class Letter(Enum):
    A = 4191
    B = 4255
    C = 4319
    D = 4383
    E = 4447
    F = 4511

    @classmethod
    def from_str(cls, let: str) -> Letter:
        match let:
            case "A":
                return Letter.A
            case "B":
                return Letter.B
            case "C":
                return Letter.C
            case "D":
                return Letter.D
            case "E":
                return Letter.E
            case "F":
                return Letter.F
            case _:
                raise ValueError("Letter is one of A to F")


@dataclass
class FiftyFourVialPlate:
    plate: Plate
    letter: Letter
    num: Num

    def value(self) -> int:
        return self.plate.value + self.letter.value + self.num.value

    @classmethod
    def from_str(cls, loc: str):
        if len(loc) != 5:
            raise ValueError("Plate locations must be PX-LY, where X is either 1 or 2 and Y is 1 to 9")
        try:
            plate = int(loc[1])
            letter = loc[3]
            num = int(loc[4])
            return FiftyFourVialPlate(plate=Plate.from_num(plate),
                                      letter=Letter.from_str(letter),
                                      num=Num.from_num(num))
        except Exception as e:
            raise ValueError("Plate locations must be PX-LY, where X is either 1 or 2 and Y is 1 to 9")

    @classmethod
    def from_int(cls, num: int) -> FiftyFourVialPlate:
        row_starts = [
            # plate 1
            FiftyFourVialPlate.from_str('P1-F1'),
            FiftyFourVialPlate.from_str('P1-E1'),
            FiftyFourVialPlate.from_str('P1-D1'),
            FiftyFourVialPlate.from_str('P1-C1'),
            FiftyFourVialPlate.from_str('P1-B1'),
            FiftyFourVialPlate.from_str('P1-A1'),
            # plate 2
            FiftyFourVialPlate.from_str('P2-F1'),
            FiftyFourVialPlate.from_str('P2-E1'),
            FiftyFourVialPlate.from_str('P2-D1'),
            FiftyFourVialPlate.from_str('P2-C1'),
            FiftyFourVialPlate.from_str('P2-B1'),
            FiftyFourVialPlate.from_str('P2-A1'),
        ]

        # find which row
        possible_row = None
        for i in range(0, 6):
            p1_val = row_starts[i].value()
            p2_val = row_starts[6 + i].value()
            if num >= p2_val:
                possible_row = row_starts[6 + i]
            elif p1_val <= num < row_starts[-1].value():
                possible_row = row_starts[i]
            if possible_row:
                break

        # determine which num
        if possible_row:
            starting_loc = possible_row
            base_val = starting_loc.plate.value + starting_loc.letter.value
            for i in range(1, 10):
                if num - i == base_val:
                    return FiftyFourVialPlate(
                        plate=starting_loc.plate,
                        letter=starting_loc.letter,
                        num=Num.from_num(i))
        raise ValueError("Number didn't match any location. " + str(num))


class TenVialColumn(Enum):
    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    SIX = 6
    SEVEN = 7
    EIGHT = 8
    NINE = 9
    TEN = 10


Tray = Union[FiftyFourVialPlate, TenVialColumn]



import os
import re
from dataclasses import dataclass
from typing import List, AnyStr, Dict

import pandas as pd
from aghplctools.ingestion.text import _no_peaks_re, _area_report_re, _header_block_re, _signal_info_re, \
    _signal_table_re, chunk_string
from result import Result, Err, Ok

from pychemstation.utils.tray_types import Tray, FiftyFourVialPlate


@dataclass
class AgilentPeak:
    peak_number: int
    retention_time: float
    peak_type: str
    width: float
    area: float
    height: float
    height_percent: float


@dataclass
class AgilentReport:
    vial_location: Tray
    signals: Dict[AnyStr, List[AgilentPeak]]
    solvents: Dict[AnyStr, AnyStr]


_column_re_dictionary = {  # regex matches for column and unit combinations
    'Peak': {  # peak index
        '#': '[ ]+(?P<Peak>[\d]+)',  # number
    },
    'RetTime': {  # retention time
        '[min]': '(?P<RetTime>[\d]+.[\d]+)',  # minutes
    },
    'Type': {  # peak type
        '': '(?P<Type>[A-Z]{1,3}(?: [A-Z]{1,2})*)',  # todo this is different from <4.8.8 aghplc tools
    },
    'Width': {  # peak width
        '[min]': '(?P<Width>[\d]+.[\d]+[e+-]*[\d]+)',
    },
    'Area': {  # peak area
        '[mAU*s]': '(?P<Area>[\d]+.[\d]+[e+-]*[\d]+)',  # area units
        '%': '(?P<percent>[\d]+.[\d]+[e+-]*[\d]+)',  # percent
    },
    'Height': {  # peak height
        '[mAU]': '(?P<Height>[\d]+.[\d]+[e+-]*[\d]+)',
    },
    'Name': {
        '': '(?P<Name>[^\s]+(?:\s[^\s]+)*)',  # peak name
    },
}


def build_peak_regex(signal_table: str) -> re.Pattern[AnyStr]:
    """
    Builds a peak regex from a signal table

    :param signal_table: block of lines associated with an area table
    :return: peak line regex object (<=3.6 _sre.SRE_PATTERN, >=3.7 re.Pattern)
    """
    split_table = signal_table.split('\n')
    if len(split_table) <= 4:  # catch peak table with no values
        return None
    # todo verify that these indicies are always true
    column_line = split_table[2]  # table column line
    unit_line = split_table[3]  # column unit line
    length_line = [len(val) + 1 for val in split_table[4].split('|')]  # length line

    # iterate over header values and units to build peak table regex
    peak_re_string = []
    for header, unit in zip(
            chunk_string(column_line, length_line),
            chunk_string(unit_line, length_line)
    ):
        if header == '':  # todo create a better catch for an undefined header
            continue
        try:
            peak_re_string.append(
                _column_re_dictionary[header][unit]  # append the appropriate regex
            )
        except KeyError:  # catch for undefined regexes (need to be built)
            raise KeyError(f'The header/unit combination "{header}" "{unit}" is not defined in the peak regex '
                           f'dictionary. Let Lars know.')
    return re.compile(
        '[ ]+'.join(peak_re_string)  # constructed string delimited by 1 or more spaces
        + '[\s]*'  # and any remaining white space
    )


# todo should be able to use the parse_area_report method of aghplctools v4.8.8

def parse_area_report(report_text: str) -> dict:
    """
    Interprets report text and parses the area report section, converting it to dictionary.

    :param report_text: plain text version of the report.
    :raises ValueError: if there are no peaks defined in the report text file
    :return: dictionary of signals in the form
        dict[wavelength][retention time (float)][Width/Area/Height/etc.]
    """
    if re.search(_no_peaks_re, report_text):  # There are no peaks in Report.txt
        raise ValueError(f'No peaks found in Report.txt')
    blocks = _header_block_re.split(report_text)
    signals = {}  # output dictionary
    for ind, block in enumerate(blocks):
        # area report block
        if _area_report_re.match(block):  # match area report block
            # break into signal blocks
            signal_blocks = _signal_table_re.split(blocks[ind + 1])
            # iterate over signal blocks
            for table in signal_blocks:
                si = _signal_info_re.match(table)
                if si is not None:
                    # some error state (e.g. 'not found')
                    if si.group('error') != '':
                        continue
                    wavelength = float(si.group('wavelength'))
                    if wavelength in signals:
                        # placeholder error raise just in case (this probably won't happen)
                        raise KeyError(
                            f'The wavelength {float(si.group("wavelength"))} is already in the signals dictionary')
                    signals[wavelength] = {}
                    # build peak regex
                    peak_re = build_peak_regex(table)
                    if peak_re is None:  # if there are no columns (empty table), continue
                        continue
                    for line in table.split('\n'):
                        peak = peak_re.match(line)
                        if peak is not None:
                            signals[wavelength][float(peak.group('RetTime'))] = {}
                            current = signals[wavelength][float(peak.group('RetTime'))]
                            for key in _column_re_dictionary:
                                if key in peak.re.groupindex:
                                    try:  # try float conversion, otherwise continue
                                        value = float(peak.group(key))
                                    except ValueError:
                                        value = peak.group(key)
                                    current[key] = value
                                else:  # ensures defined
                                    current[key] = None
    return signals


def process_export_report(file_path, target_wavelengths=None, min_retention_time=0, max_retention_time=999):
    # # Pull signals from the file
    # from aghplctools.ingestion.text import pull_hplc_area_from_txt
    # signals = pull_hplc_area_from_txt(file_path)

    with open(file_path, 'r', encoding='utf-16') as openfile:
        text = openfile.read()

    try:
        signals = parse_area_report(text)
    except ValueError as e:
        # value error thrown if there are no peaks found in the report
        print(e)
        return [], [], []

    # filter wavelengths by the ones to keep
    if target_wavelengths is not None:
        signals = {key: signals[key] for key in target_wavelengths if key in signals}

    wavelengths = []
    retention_times = []
    areas = []

    for wavelength, wavelength_dict in signals.items():
        for ret_time, ret_time_dict in wavelength_dict.items():
            if min_retention_time <= ret_time <= max_retention_time:
                wavelengths.append(wavelength)
                retention_times.append(ret_time)
                areas.append(ret_time_dict['Area'])

    return wavelengths, retention_times, areas


def process_folder(folder_path, target_wavelengths=None, min_retention_time=0, max_retention_time=999):
    # folder path is the path to the overall folder, and inside there should be subfolders for each LC sample
    # each subfolder should have a Report.TXT file
    # sample_names = []
    wavelengths = []
    retention_times = []
    peak_areas = []

    # Get a list of all items (files and directories) in the folder
    items = [os.path.join(folder_path, item) for item in os.listdir(folder_path)]

    # Filter only directories from the list
    # folders = [item for item in items if os.path.isdir(item)]

    # # Sort the folders by creation date
    # sorted_folders = sorted(folders, key=lambda f: os.stat(f).st_ctime)

    for filename in items:
        if filename.endswith('Report.TXT'):
            # file_path = os.path.join(subfolder, filename)
            file_wavelengths, file_retention_times, file_peak_areas = process_export_report(filename,
                                                                                            target_wavelengths,
                                                                                            min_retention_time,
                                                                                            max_retention_time)
            wavelengths.extend(file_wavelengths)
            retention_times.extend(file_retention_times)
            peak_areas.extend(file_peak_areas)

    results_df = pd.DataFrame({'Wavelengths': wavelengths, 'Retention Times': retention_times, 'Areas': peak_areas})

    # Save the results to a CSV file
    # results_csv_path = os.path.join(folder_path, 'all_sample_data.csv')  # Specify the desired file path
    # results_df.to_csv(results_csv_path, index=False)
    # print(f"Results saved to {results_csv_path}")
    return results_df


def process_csv_report(folder_path: str) -> Result[AgilentReport, AnyStr]:
    labels = os.path.join(folder_path, f'REPORT00.CSV')
    if os.path.exists(labels):
        df_labels: Dict[int, Dict[int: AnyStr]] = pd.read_csv(labels, encoding="utf-16", header=None).to_dict()
        vial_location = []
        signals = {}
        solvents = {}
        for pos, val in df_labels[0].items():
            if val == "Location":
                vial_location = df_labels[1][pos]
            elif "Solvent" in val:
                if val not in solvents.keys():
                    solvents[val] = df_labels[2][pos]
            elif val == "Number of Signals":
                num_signals = int(df_labels[1][pos])
                for s in range(1, num_signals + 1):
                    peaks = process_peaks(os.path.join(folder_path, f'REPORT0{s}.CSV'))
                    if peaks.is_ok():
                        wavelength = df_labels[1][pos + s].partition(",4 Ref=off")[0][-3:]
                        signals[wavelength] = peaks.ok_value
                break

        return Ok(AgilentReport(
            signals=signals,
            vial_location=FiftyFourVialPlate.from_int(vial_location),
            solvents=solvents
        ))

    return Err("No report found")


def process_peaks(folder_path: str) -> Result[List[AgilentPeak], AnyStr]:
    try:
        df = pd.read_csv(folder_path, encoding="utf-16", header=None)
        return Ok(df.apply(lambda row: AgilentPeak(*row), axis=1))
    except Exception:
        return Err("Trouble reading report")



from ....control.controllers import CommunicationController
from .device import DeviceController
from ....utils.pump_types import Pump
from ....utils.table_types import Table


class PumpController(DeviceController):

    def __init__(self, controller: CommunicationController, table: Table):
        super().__init__(controller, table)
        self.A1 = Pump(in_use=True, solvent="A1")
        self.B1 = Pump(in_use=True, solvent="B1")
        self.A2 = Pump(in_use=False, solvent="A2")
        self.B2 = Pump(in_use=False, solvent="B2")

    def validate_pumps(self):
        invalid_A_pump_usage = self.A1.in_use and self.A2.in_use
        invalid_B_pump_usage = self.B1.in_use and self.B2.in_use
        if invalid_A_pump_usage or invalid_B_pump_usage:
            raise AttributeError

    def switch_pump(self, num: int, pump: str):
        if pump == "A":
            if num == 1:
                self.A1.in_use = True
                self.A2.in_use = False
            elif num == 2:
                self.A1.in_use = False
                self.A2.in_use = True
        elif pump == "B":
            if num == 1:
                self.B1.in_use = True
                self.B2.in_use = False
            elif num == 2:
                self.B1.in_use = False
                self.B2.in_use = True
        self.purge()

    def purge(self):
        pass

    def get_row(self, row: int):
        pass



from __future__ import annotations

from enum import Enum
from typing import Union
from dataclasses import dataclass


@dataclass
class Response:
    string_response: str
    num_response: Union[int, float]


# Commands sent to the Chemstation Macro
# See https://www.agilent.com/cs/library/usermanuals/Public/MACROS.PDF
class Command(Enum):
    def __str__(self):
        return '%s' % self.value

    GET_NUM_VAL_CMD = "response_num = {cmd}"
    GET_TEXT_VAL_CMD = "response$ = {cmd}"
    RESET_COUNTER_CMD = "last_cmd_no = 0"
    GET_STATUS_CMD = "response$ = AcqStatus$"
    SLEEP_CMD = "Sleep {seconds}"
    STANDBY_CMD = "Standby"
    STOP_MACRO_CMD = "Stop"
    PREPRUN_CMD = "PrepRun"

    # Instrument Control
    LAMP_ON_CMD = "LampAll ON"
    LAMP_OFF_CMD = "LampAll OFF"
    PUMP_ON_CMD = "PumpAll ON"
    PUMP_OFF_CMD = "PumpAll OFF"
    INSTRUMENT_OFF = 'macro "SHUTDOWN.MAC" ,go'
    INSTRUMENT_ON = 'LIDoOperation "TURN_ON"'

    # Method and Sequence Related
    GET_METHOD_CMD = "response$ = _MethFile$"
    GET_ROWS_CMD = 'response_num = TabHdrVal({register}, "{table_name}", "{col_name}")'
    SWITCH_METHOD_CMD = 'LoadMethod _MethPath$, _MethFile$'
    SWITCH_METHOD_CMD_SPECIFIC = 'LoadMethod "{method_dir}", "{method_name}.M"'
    START_METHOD_CMD = "StartMethod"
    RUN_METHOD_CMD = 'RunMethod "{data_dir}",, "{experiment_name}_{timestamp}"'
    STOP_METHOD_CMD = "StopMethod"
    UPDATE_METHOD_CMD = 'UpdateMethod'
    SWITCH_SEQUENCE_CMD = 'LoadSequence _SeqPath$, _SeqFile$'
    SAVE_SEQUENCE_CMD = 'SaveSequence _SeqPath$, _SeqFile$'
    SAVE_METHOD_CMD = 'SaveMethod _MethPath$, _MethFile$, "{commit_msg}"'
    GET_SEQUENCE_CMD = 'response$ = _SeqFile$'
    RUN_SEQUENCE_CMD = 'RunSequence'


class HPLCRunningStatus(Enum):
    @classmethod
    def has_member_key(cls, key):
        return key in cls.__members__

    INJECTING = "INJECTING"
    PREPARING = "PREPARING"
    RUN = "RUN"
    NOTREADY = "NOTREADY"
    POSTRUN = "POSTRUN"
    RAWDATA = "RAWDATA"
    INITIALIZING = "INITIALIZING"
    NOMODULE = "NOMODULE"


class HPLCAvailStatus(Enum):
    @classmethod
    def has_member_key(cls, key):
        return key in cls.__members__

    PRERUN = "PRERUN"
    OFFLINE = "OFFLINE"
    STANDBY = "STANDBY"


class HPLCErrorStatus(Enum):

    @classmethod
    def has_member_key(cls, key):
        return key in cls.__members__

    ERROR = "ERROR"
    BREAK = "BREAK"
    NORESPONSE = "NORESPONSE"
    MALFORMED = "MALFORMED"


def str_to_status(status: str) -> Union[HPLCAvailStatus, HPLCErrorStatus, HPLCRunningStatus]:
    if HPLCErrorStatus.has_member_key(status):
        return HPLCErrorStatus[status]
    if HPLCRunningStatus.has_member_key(status):
        return HPLCRunningStatus[status]
    if HPLCAvailStatus.has_member_key(status):
        return HPLCAvailStatus[status]
    raise KeyError(status)


Status = Union[HPLCRunningStatus, HPLCAvailStatus, HPLCErrorStatus]



from ....control.controllers import CommunicationController
from .device import DeviceController
from ....utils.injector_types import *
from ....utils.macro import Command
from ....utils.table_types import Table, RegisterFlag


class InjectorController(DeviceController):

    def __init__(self, controller: CommunicationController, table: Table, offline: bool):
        super().__init__(controller, table, offline)

    def get_row(self, row: int) -> InjectorFunction:
        def return_tray_loc() -> Tray:
            pass

        function = self.get_text(row, RegisterFlag.FUNCTION)
        if function == "Wait":
            return Wait(duration=self.get_num(row, RegisterFlag.TIME))
        elif function == "Inject":
            return Inject()
        elif function == "Draw":
            # TODO: better error handling
            is_source = SourceType(self.get_text(row, RegisterFlag.DRAW_SOURCE))
            is_volume = Mode(self.get_text(row, RegisterFlag.DRAW_VOLUME))
            vol = self.get_num(row, RegisterFlag.DRAW_VOLUME_VALUE) if is_volume == Mode.SET else None
            if is_source is SourceType.SPECIFIC_LOCATION:
                return Draw(amount=vol, source=return_tray_loc())
            elif is_source is SourceType.LOCATION:
                return Draw(amount=vol, location=self.get_text(row, RegisterFlag.DRAW_LOCATION))
        elif function == "Remote":
            return Remote(command=RemoteCommand(self.get_text(row, RegisterFlag.REMOTE)),
                          duration=self.get_num(row, RegisterFlag.REMOTE_DUR))

    def load(self) -> InjectorTable:
        rows = self.get_num_rows()
        if rows.is_ok():
            return InjectorTable(functions=[self.get_row(i) for i in range(int(rows.ok_value.num_response))])

    def edit(self, injector_table: InjectorTable):
        columns_added = set()

        def add_table_val(col_name: RegisterFlag, val: Union[str, int, float]):
            nonlocal columns_added
            if True:
                if isinstance(val, str):
                    self.edit_row_text(col_name=col_name, val=val)
                else:
                    self.edit_row_num(col_name=col_name, val=val)
            else:
                if isinstance(val, str):
                    self.add_new_col_text(col_name=col_name, val=val)
                else:
                    self.add_new_col_num(col_name=col_name, val=val)
                columns_added.add(col_name)

        def add_inject(inject: Inject):
            add_table_val(col_name=RegisterFlag.FUNCTION, val=inject.__class__.__name__)

        def add_draw(draw: Draw):
            add_table_val(col_name=RegisterFlag.FUNCTION, val=draw.__class__.__name__)
            add_table_val(col_name=RegisterFlag.DRAW_SPEED, val=SourceType.DEFAULT.value)
            add_table_val(col_name=RegisterFlag.DRAW_OFFSET, val=SourceType.DEFAULT.value)

            if draw.amount:
                add_table_val(col_name=RegisterFlag.DRAW_VOLUME, val=Mode.SET.value)
                add_table_val(col_name=RegisterFlag.DRAW_VOLUME_VALUE, val=draw.amount)
            else:
                add_table_val(col_name=RegisterFlag.DRAW_VOLUME, val=Mode.DEFAULT.value)

            if draw.location:
                add_table_val(col_name=RegisterFlag.DRAW_SOURCE, val=SourceType.LOCATION.value)
                add_table_val(col_name=RegisterFlag.DRAW_LOCATION, val=draw.location)
            elif draw.source:
                add_table_val(col_name=RegisterFlag.DRAW_SOURCE, val=SourceType.SPECIFIC_LOCATION.value)
                add_table_val(col_name=RegisterFlag.DRAW_LOCATION_UNIT, val=1)
                add_table_val(col_name=RegisterFlag.DRAW_LOCATION_TRAY, val=1)
                add_table_val(col_name=RegisterFlag.DRAW_LOCATION_ROW, val=1)
                add_table_val(col_name=RegisterFlag.DRAW_LOCATION_COLUMN, val=1)
            else:
                add_table_val(col_name=RegisterFlag.DRAW_SOURCE, val=SourceType.DEFAULT.value)

        def add_wait(wait: Wait):
            add_table_val(col_name=RegisterFlag.FUNCTION, val=wait.__class__.__name__)
            add_table_val(col_name=RegisterFlag.TIME, val=wait.duration)

        def add_remote(remote: Remote):
            add_table_val(col_name=RegisterFlag.FUNCTION, val=remote.__class__.__name__)
            add_table_val(col_name=RegisterFlag.REMOTE, val=remote.command.value)
            add_table_val(col_name=RegisterFlag.REMOTE_DUR, val=remote.duration)

        self.send(Command.SAVE_METHOD_CMD)
        rows = self.get_num_rows()
        if rows.is_ok():
            existing_row_num = rows.value.num_response
            for i, function in enumerate(injector_table.functions):
                if (i+1) > existing_row_num:
                    self.add_row()
                if isinstance(function, Inject):
                    add_inject(function)
                elif isinstance(function, Draw):
                    add_draw(function)
                elif isinstance(function, Wait):
                    add_wait(function)
                elif isinstance(function, Remote):
                    add_remote(function)
                self.download()
                self.send(Command.SAVE_METHOD_CMD)
                self.send(Command.SWITCH_METHOD_CMD)
                existing_row_num = self.get_num_rows().ok_value.num_response

    def download(self):
        self.send('Sleep 1')
        self.sleepy_send("DownloadRCMethod WLS1")
        self.send('Sleep 1')
        self.sleepy_send("DownloadLWls 1")
        self.send('Sleep 1')


"""
Module to provide API for higher-level HPLC actions.

Authors: Lucy Hao
"""

from typing import Union, Optional

from .controllers.devices.injector import InjectorController
from ..control.controllers import MethodController, SequenceController, CommunicationController
from ..utils.chromatogram import AgilentChannelChromatogramData
from ..utils.injector_types import InjectorTable
from ..utils.macro import Command, HPLCRunningStatus, HPLCAvailStatus, HPLCErrorStatus, Response
from ..utils.method_types import MethodDetails
from ..utils.sequence_types import SequenceTable, SequenceEntry
from ..utils.table_types import Table


class HPLCController:
    # tables
    METHOD_TIMETABLE = Table(
        register="RCPMP1Method[1]",
        name="Timetable"
    )

    SEQUENCE_TABLE = Table(
        register="_sequence[1]",
        name="SeqTable1"
    )

    INJECTOR_TABLE = Table(
        register="RCWLS1Pretreatment[1]",
        name="InstructionTable"
    )

    MSD_TABLE = Table(
        register="MSACQINFO[1]",
        name="SprayChamber"
    )

    def __init__(self,
                 comm_dir: str,
                 data_dir: str,
                 method_dir: str,
                 sequence_dir: str,
                 sequence_data_dir: Optional[str] = None,
                 offline: bool = False):
        """Initialize HPLC controller. The `hplc_talk.mac` macro file must be loaded in the Chemstation software.
        `comm_dir` must match the file path in the macro file.

        :param comm_dir: Name of directory for communication, where ChemStation will read and write from. Can be any existing directory.
        :param data_dir: Name of directory for storing data after method or sequence runs.
        :param method_dir: Name of directory where method files are stored.
        :param sequence_dir: Name of directory where sequence files are stored.
        :param sequence_data_dir: Name of directory for storing sequence data, if different from method
        :raises FileNotFoundError: If either `data_dir`, `method_dir`, `sequence_dir`, `sequence_data_dir`or `comm_dir` is not a valid directory.
        """
        self.comm = CommunicationController(comm_dir=comm_dir) if not offline else None
        self.method_controller = MethodController(controller=self.comm,
                                                  src=method_dir,
                                                  data_dir=data_dir,
                                                  table=self.METHOD_TIMETABLE,
                                                  offline=offline,
                                                  injector_controller=InjectorController(controller=self.comm,
                                                                                         table=self.INJECTOR_TABLE,
                                                                                         offline=offline))
        self.sequence_controller = SequenceController(controller=self.comm,
                                                      src=sequence_dir,
                                                      data_dir=data_dir if not sequence_data_dir else sequence_data_dir,
                                                      table=self.SEQUENCE_TABLE,
                                                      method_dir=method_dir,
                                                      offline=offline)

    def send(self, cmd: Union[Command, str]):
        if not self.comm:
            raise RuntimeError(
                "Communication controller must be initialized before sending command. It is currently in offline mode.")
        self.comm.send(cmd)

    def receive(self) -> Response:
        if not self.comm:
            raise RuntimeError(
                "Communication controller must be initialized before sending command. It is currently in offline mode.")
        return self.comm.receive().value

    def status(self) -> Union[HPLCRunningStatus | HPLCAvailStatus | HPLCErrorStatus]:
        if not self.comm:
            raise RuntimeError(
                "Communication controller must be initialized before sending command. It is currently in offline mode.")
        return self.comm.get_status()

    def switch_method(self, method_name: str):
        """
        Allows the user to switch between pre-programmed methods. No need to append '.M'
        to the end of the method name. For example. for the method named 'General-Poroshell.M',
        only 'General-Poroshell' is needed.

        :param method_name: any available method in Chemstation method directory
        :raises  IndexError: Response did not have expected format. Try again.
        :raises  AssertionError: The desired method is not selected. Try again.
        """
        self.method_controller.switch(method_name)

    def switch_sequence(self, sequence_name: str):
        """
         Allows the user to switch between pre-programmed sequences. The sequence name does not need the '.S' extension.
         For example. for the method named 'mySeq.S', only 'mySeq' is needed.

        :param sequence_name: The name of the sequence file
        """
        self.sequence_controller.switch(sequence_name)

    def run_method(self, experiment_name: str, stall_while_running: bool = True):
        """
        This is the preferred method to trigger a run.
        Starts the currently selected method, storing data
        under the <data_dir>/<experiment_name>.D folder.
        The <experiment_name> will be appended with a timestamp in the '%Y-%m-%d-%H-%M' format.
        Device must be ready.

        :param experiment_name: Name of the experiment
        :param stall_while_running: whether this method should return or stall while HPLC runs.
        """
        self.method_controller.run(experiment_name, stall_while_running)

    def run_sequence(self, stall_while_running: bool = True):
        """
        Starts the currently loaded sequence, storing data
        under the <data_dir>/<sequence table name> folder.
        Device must be ready.

        :param stall_while_running: whether this method should return or stall while HPLC runs.
        """
        self.sequence_controller.run(stall_while_running=stall_while_running)

    def edit_method(self, updated_method: MethodDetails, save: bool = False):
        """Updated the currently loaded method in ChemStation with provided values.

        :param updated_method: the method with updated values, to be sent to Chemstation to modify the currently loaded method.
        :param save: whether this method should be to disk, or just modified.
        """
        self.method_controller.edit(updated_method, save)

    def edit_sequence(self, updated_sequence: SequenceTable):
        """
        Updates the currently loaded sequence table with the provided table.
        If you would only like to edit a single row of a sequence table, use `edit_sequence_row` instead.

        :param updated_sequence: The sequence table to be written to the currently loaded sequence table.
        """
        self.sequence_controller.edit(updated_sequence)

    def edit_sequence_row(self, row: SequenceEntry, num: int):
        """
        Edits a row in the sequence table. Assumes the row already exists.

        :param row: sequence row entry with updated information
        :param num: the row to edit, based on 1-based indexing
        """
        self.sequence_controller.edit_row(row, num)

    def get_last_run_method_data(self, read_uv: bool = False,
                                 data: Optional[str] = None) -> AgilentChannelChromatogramData:
        """
        Returns the last run method data.

        :param data: If you want to just load method data but from a file path. This file path must be the complete file path.
        :param read_uv: whether to also read the UV file
        """
        return self.method_controller.get_data(custom_path=data, read_uv=read_uv)

    def get_last_run_sequence_data(self, read_uv: bool = False,
                                   data: Optional[str] = None) -> list[AgilentChannelChromatogramData]:
        """
        Returns data for all rows in the last run sequence data.

        :param data: If you want to just load sequence data but from a file path. This file path must be the complete file path.
        :param read_uv: whether to also read the UV file
        """
        return self.sequence_controller.get_data(custom_path=data, read_uv=read_uv)

    def check_loaded_sequence(self) -> str:
        """
        Returns the name of the currently loaded sequence.
        """
        return self.sequence_controller.check()

    def check_loaded_method(self) -> str:
        """
        Returns the name of the currently loaded method.
        """
        return self.method_controller.check()

    def load_injector_program(self) -> InjectorTable:
        return self.method_controller.injector_controller.load()

    def load_method(self) -> MethodDetails:
        """
        Returns all details of the currently loaded method, including its timetable.
        """
        return self.method_controller.load()

    def load_sequence(self) -> SequenceTable:
        """
        Returns the currently loaded sequence.
        """
        return self.sequence_controller.load()

    def standby(self):
        """Switches all modules in standby mode. All lamps and pumps are switched off."""
        self.send(Command.STANDBY_CMD)

    def preprun(self):
        """ Prepares all modules for run. All lamps and pumps are switched on."""
        self.send(Command.PREPRUN_CMD)

    def lamp_on(self):
        """Turns the UV lamp on."""
        self.send(Command.LAMP_ON_CMD)

    def lamp_off(self):
        """Turns the UV lamp off."""
        self.send(Command.LAMP_OFF_CMD)

    def pump_on(self):
        """Turns on the pump on."""
        self.send(Command.PUMP_ON_CMD)

    def pump_off(self):
        """Turns the pump off."""
        self.send(Command.PUMP_OFF_CMD)

    def instrument_off(self):
        """Shuts the entire instrument off, including pumps, lamps, thermostat."""
        self.send(Command.INSTRUMENT_OFF)

    def instrument_on(self):
        """Turns the entire instrument on, including pumps, lamps, thermostat."""
        self.send(Command.INSTRUMENT_ON)
